import copy
import json
import os
import re
import subprocess
import time
from dataclasses import dataclass
from html import unescape
from typing import Dict, Generator
from xml.etree import ElementTree

import jwt
import ldap
import requests
from requests.cookies import RequestsCookieJar
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.chrome.webdriver import WebDriver
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from tenacity import retry, stop_after_attempt, wait_fixed

from ucsschool.lib.models.utils import ucr
from univention.admin.uldap import access

SELF_DISCLOSURE_URL = f"https://self-disclosure1.{ucr['domainname']}/ucsschool/apis/self_disclosure/v1"
TEST_APP_URL = f"https://kc1.{ucr['domainname']}/univention-test-app/"


def get_ldap_base_from_traeger(hostname: str) -> str:
    ldap_con = ldap.initialize(f"ldap://{hostname}:7389")
    ncs = ldap_con.get_naming_contexts()
    ncs.remove(b"cn=translog")
    return ncs[0].decode("UTF-8")


def get_traeger_ldap_con(traeger: str) -> access:
    """
    UCS_ENV_TRAEGER1_DOMAIN and UCS_ENV_TRAEGER2_DOMAIN
    needs to be set by the (jenkins) job
    dito for the /var/lib/ucs/traeger1[2].secret files
    """
    traegers = {
        "Traeger1": f"ucs-sso.{os.environ.get('UCS_ENV_TRAEGER1_DOMAIN')}",
        "Traeger2": f"ucs-sso.{os.environ.get('UCS_ENV_TRAEGER2_DOMAIN')}",
    }
    hostname = traegers[traeger]
    password = open(f"/var/lib/ucs-test/{traeger}.secret").read().strip()
    ldap_base = get_ldap_base_from_traeger(hostname)
    lo = access(
        hostname,
        start_tls=1,
        base=ldap_base,
        binddn=f"uid=Administrator,cn=users,{ldap_base}",
        bindpw=password,
    )
    return lo


def get_traeger_ldap_admin_con(traeger: str) -> access:
    """
    UCS_ENV_TRAEGER1_DOMAIN and UCS_ENV_TRAEGER2_DOMAIN
    needs to be set by the (jenkins) job
    dito for the /var/lib/ucs/traeger1[2].secret files
    """
    traegers = {
        "Traeger1": f"ucs-sso.{os.environ.get('UCS_ENV_TRAEGER1_DOMAIN')}",
        "Traeger2": f"ucs-sso.{os.environ.get('UCS_ENV_TRAEGER2_DOMAIN')}",
    }
    hostname = traegers[traeger]
    password = (
        subprocess.check_output(
            [
                "univention-ssh",
                f"/var/lib/ucs-test/{traeger}.secret",
                f"Administrator@{hostname}",
                "cat",
                "/etc/ldap.secret",
            ]
        )
        .decode()
        .strip()
    )
    ldap_base = get_ldap_base_from_traeger(hostname)
    lo = access(
        hostname,
        start_tls=1,
        base=ldap_base,
        binddn=f"cn=admin,{ldap_base}",
        bindpw=password,
    )
    return lo


@dataclass
class User:
    username: str
    firstname: str
    lastname: str
    displayName: str
    ucsschoolRole: list
    dn: str
    entryUUID: str


def get_user_from_traeger(traeger: str, filter_l: str) -> Generator[User, None, None]:
    for dn, attrs in get_traeger_ldap_con(traeger).search(filter_l, attr=["*", "+"]):
        yield User(
            username=attrs.get("uid", [b""])[0].decode("utf-8"),
            firstname=attrs.get("givenName", [b""])[0].decode("utf-8"),
            lastname=attrs.get("sn", [b""])[0].decode("utf-8"),
            displayName=attrs.get("gecos", [b""])[0].decode("utf-8"),
            ucsschoolRole=[role.decode("utf-8") for role in attrs.get("ucsschoolRole", [b""])],
            dn=dn,
            entryUUID=attrs.get("entryUUID", [b""])[0].decode("utf-8"),
        )


@retry(wait=wait_fixed(5), stop=stop_after_attempt(6))
def get_self_disclosure_metadata(access_token: str) -> Dict:
    sub = jwt.decode(access_token, verify=False)["sub"]
    uri = f"{SELF_DISCLOSURE_URL}/users/{sub}/metadata"
    headers = {"Authorization": f"Bearer {access_token}"}
    response = requests.get(uri, headers=headers)
    response.raise_for_status()
    return json.loads(response.text)


def get_self_disclosure_user_groups(access_token: str) -> Dict:
    sub = jwt.decode(access_token, verify=False)["sub"]
    uri = f"{SELF_DISCLOSURE_URL}/users/{sub}/groups"
    headers = {"Authorization": f"Bearer {access_token}"}
    response = requests.get(uri, headers=headers)
    response.raise_for_status()
    return json.loads(response.text)


def get_self_disclosure_group_users(access_token: str, group_id: str) -> Dict:
    uri = f"{SELF_DISCLOSURE_URL}/groups/{group_id}/users"
    headers = {"Authorization": f"Bearer {access_token}"}
    response = requests.get(uri, headers=headers)
    response.raise_for_status()
    return response.json()


def wait_for(driver: WebDriver, by: By, element: str, timeout: int) -> None:
    element_present = EC.presence_of_element_located((by, element))
    WebDriverWait(driver, timeout).until(element_present)
    WebDriverWait(driver, timeout).until(EC.visibility_of_element_located((by, element)))
    time.sleep(1)


def wait_for_id(driver: WebDriver, element_id: str, timeout: int = 5) -> WebElement:
    wait_for(driver, By.ID, element_id, timeout)
    return driver.find_element_by_id(element_id)


def get_token_from_test_app_logon(
    selenium: WebDriver, username: str, password: str, kc_idp_hint: str
) -> Dict:
    selenium.get(TEST_APP_URL)
    wait_for_id(selenium, "login")
    selenium.find_element_by_name("kc_idp_hint").send_keys(kc_idp_hint)
    selenium.find_element_by_name("pkce").send_keys("y")
    selenium.find_element_by_id("login").click()
    wait_for_id(selenium, "umcLoginDialog")
    selenium.find_element_by_id("umcLoginUsername").send_keys(username)
    selenium.find_element_by_id("umcLoginPassword").send_keys(password)
    try:
        # UCS < 5.0
        selenium.find_element_by_id("umcLoginSubmit").click()
    except NoSuchElementException:
        # UCS >= 5.0
        selenium.find_element_by_name("submit").click()
    wait_for_id(selenium, "pseudonym")
    tokens_json = selenium.find_element_by_tag_name("pre").text
    return json.loads(tokens_json.replace("'", '"'))


def get_xpath_element(xpath, xml, default=None):
    parsed_page = ElementTree.fromstring(bytes(xml, encoding="utf8"))
    elem = parsed_page.find(".//{http://www.w3.org/1999/xhtml}%s" % (xpath,))
    if elem is None:
        elem = parsed_page.find(".//%s" % (xpath,))
    if elem is None:
        elem = default
    return elem


def ucs_saml_login(username: str, password: str, fqdn: str) -> RequestsCookieJar:
    with requests.Session() as session:
        url = f"https://{fqdn}/univention/saml/"
        response = session.get(url, verify=False)
        auth_state = get_xpath_element('input[@name="AuthState"]', response.text).get("value")
        data = {"username": username, "password": password, "AuthState": auth_state}
        response = session.post(response.url, data=data, allow_redirects=True)
        assert "SimpleSAMLAuthToken" in session.cookies.keys(), session.cookies.keys()
        return session.cookies


def univention_test_app_login_with_saml(traeger: str, fqdn: str, cookies: RequestsCookieJar) -> str:
    with requests.Session() as session:
        session.cookies = copy.deepcopy(cookies)
        url = f"https://{fqdn}/univention-test-app/login"
        params = {"kc_idp_hint": traeger, "pkce": "y"}
        # 1. univention-test-app/login
        response = session.get(url, params=params, verify=False, allow_redirects=True)
        # no valid xml :-(
        # sp_url = get_xpath_element('form[@method="post"]', response.text).get('action')
        url = re.search(r'action="([^"]+)"', response.text).groups()[0]
        saml_request = re.search(r'name="SAMLRequest" value="([^"]+)"', response.text).groups()[0]
        saml_relay_state = re.search(r'name="RelayState" value="([^"]+)"', response.text).groups()[0]
        data = {"SAMLRequest": saml_request, "RelayState": saml_relay_state}
        # 2. simplesamlphp/saml2/idp/SSOService.php
        response = session.post(url, data=data, verify=False)
        url = get_xpath_element('form[@method="post"]', response.text).get("action")
        saml_response = get_xpath_element('input[@name="SAMLResponse"]', response.text).get("value")
        saml_relay_state = get_xpath_element('input[@name="RelayState"]', response.text).get("value")
        data = {"SAMLResponse": saml_response, "RelayState": saml_relay_state}
        # 3. auth/realms/ID-Broker/broker/traeger1/endpoint
        response = session.post(url, data=data, verify=False, allow_redirects=True)
        assert response.status_code == 200
        access_token = re.search(r">{'access_token': '([^']+)'", unescape(response.text)).groups()[0]
        return access_token
